// Copyright 2017, James Sarrett.
// Copyright 2017, Bastiaan Olij.
// SPDX-License-Identifier: BSL-1.0
/*
 * OpenHMD - Free and Open Source API and drivers for immersive technology.
 */

/* Shaders */


#ifndef SHADERS_H
#define SHADERS_H

extern const char * const distortion_vert;
extern const char * const distortion_frag;

extern const char * const distortion_vert_330;
extern const char * const distortion_frag_330;

extern const char * const distortion_vert_es;
extern const char * const distortion_frag_es;

#endif /* SHADERS_H */
